/* linuxrc.c
 * Copyright Rene Mayrhofer <rmayr@vianova.at>, 1999
 *
 * I hereby put this program under the terms of the Artistic license.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/mount.h>

#define CAPTURE_STDOUT "/tmp/stdout"
#define CAPTURE_STDERR "/tmp/stderr"

#define MAX_MODULENAME_LENGTH 15			/* the module name */
#define MAX_MODULESTRING_LENGTH 80		/* the module name with parameters */
#define MAX_CMDLINE_LENGTH 256				/* the length of the kernel command line */
#define MAX_ROOT_DEV_LENGTH 20				/* the length of a device name */
#define KERNEL_REAL_ROOT_DEV_FILE "/proc/sys/kernel/real-root-dev"
#define KERNEL_CMDLINE "/proc/cmdline"

int exec_insmod(char* modulestr);
int sys(const char *command, const char* parameter,
					const char *stdout_target, const char *stdout_mode,
					const char *stderr_target, const char *stderr_mode);
int get_real_root_dev();
int set_real_root_dev(const unsigned int root_dev);
int get_root_from_cmdline(char *root_dev);

int main(int argc, char **argv, char **env)
{
  char *boot_modules, *real_root_dev, /*root_dev[MAX_CMDLINE_LENGTH],*/
  			module[80];
  int i=0, len, j, root_dev_nr;

	printf("\nInitrd boot modules loader starting. "
	       "[Rene Mayrhofer <rmayr@vianova.at>, 1999]\n");

	boot_modules = getenv("boot_modules");
	if (boot_modules) {
    /* set the path where insmod searches for module files */
    setenv("MODPATH", "/lib", 1);

    /* set output streams to be unbuffered */
    setvbuf(stdout, (char*) NULL, _IONBF, 0);
    setvbuf(stderr, (char*) NULL, _IONBF, 0);
    fflush(stdout);
    fflush(stderr);

    /* parse the boot_modules string */
    len=strlen(boot_modules);
    while(i < len) {
      /* get the modules to load (separated by commas) */
      j=0;
      while(i < len && boot_modules[i] != ',') {
        module[j] = boot_modules[i];
				i++; j++;
      }
      module[j] = 0;
      if(j > 0) {
				exec_insmod(module);
      }
      i++;
    }
  } else {
  	/* no modules to load */
  }
  fputs("\n", stdout);

  /* first mount the proc dev to be able to manipulate the kernel */
  if(mount("proc", "/proc", "proc", 0, 0) == 0) {
   	/* if the real root device was given on the kernel command line, use it */
   	/* This stays in for users with very special setups, when the root
   	   device can not be detected by the kernel and can not be specified
   	   by a 'root=/dev/xxxxxx' paramater on the commandline. */
	  real_root_dev = getenv("real_root_dev");
		if(real_root_dev == NULL) {
			/* the real_root_dev was not given directly */
  		/* check if the real root device was detected by the kernel
  		   during bootup */
	  	if((root_dev_nr = get_real_root_dev()) <= 0) {
  			
  			printf("The root device was not detected by the kernel.\n");
		
  			/* what should be done here ??? */
  			
		    /* set the real root device (needed if the root dev was not detected
  			   by the kernel during initial bootup) */
/*	  		if(get_root_from_cmdline(root_dev) == 0) {
	  			printf("Root device as set on the kernel commandline: %s\n", root_dev);
				}*/
	  	} else {
	  	  /* ok, the root device has already been set by the kernel */
  		}
  		
  	} else {
			/* use the value given by the special command line variable */
    	printf("Setting kernel root device to %s .. ", real_root_dev);
    	if(set_real_root_dev((unsigned int) atoi(real_root_dev)) == 0)
		    printf("successful.\n");
		  else
		  	printf("unsuccessful.\n");
  	}
  	
	  if(umount("/proc") != 0) {
  	  printf("Unable to unmount the proc filesystem !\n");
	  }
	
	} else {
    printf("Unable to mount the proc filesystem !\n");
    exit(-1);
  }

  fputs("\n",stdout);
  return 0;
}

int exec_insmod(char *modulestr) {
  FILE *read_stderr;
  int i=0, len, ret, ch, name_ended=0, name_len=0;
  char modulename[MAX_MODULENAME_LENGTH];
   
  len=strlen(modulestr);
  /* get the module parameters (separated by ':' */
  while(i < len) {
    if(modulestr[i] == ':') {
      modulestr[i] = ' ';
      if(! name_ended)
        name_ended = 1;
    }
    else {
      if(! name_ended) {
        modulename[i] = modulestr[i];
        name_len++;
      }
    }
    i++;
  }
  modulename[name_len] = 0;
  printf("loading module: %-*s .... ", MAX_MODULENAME_LENGTH, modulename);
  
  /* try to execute insmod */
	ret = sys("/sbin/insmod", modulestr, CAPTURE_STDOUT, "a", CAPTURE_STDERR, "w");
   
  if(ret == 0) {
    printf("successfully loaded.\n");
    return 0;
  } else {
    printf("unsuccessful ! The error was:\n");
    read_stderr = fopen(CAPTURE_STDERR, "r");
    while( (ch = fgetc(read_stderr)) != EOF)
      putc(ch, stdout);
    fclose(read_stderr);
    printf("\n");
    return -2;
  }
  
  return -3;
}


/* mainly taken from "man 3 system" */
int sys(const char *command, const char *parameter,
					const char *stdout_target, const char *stdout_mode,
					const char *stderr_target, const char *stderr_mode) {
	int pid, status;
	char *argv[4];

	if(command == 0)
		return 1;
	/* start a new process */
	pid = fork();
	if(pid == -1)
		return -1;
	if(pid == 0) {
		argv[0] = command;
		argv[1] = parameter;
		argv[2] = 0;
		
		/* set the output places */
	  /* (set stdout and stderr to temporary files to capture the output) */
	  if( (stdout = freopen(stdout_target, stdout_mode, stdout)) == NULL) {
  	  return -2;
	  }
	  if( (stderr = freopen(stderr_target, stderr_mode, stderr)) == NULL) {
	    return -2;
	  }
		
		/* execute */
		execv(command, argv);
		exit(127);
	}
	
	/* wait for completion */
	do {
		if(waitpid(pid, &status, 0) == -1) {
			if(errno != EINTR) {
				return -1;
			}
		} else
			return status;
	} while(1);
}

/* get_real_root_dev returns a value < 0 on error.
 * A value >= 0 specifies the real root device.
 */
int get_real_root_dev() {
  FILE *kernel_root_dev;
	char str_root_dev[5];

  if( (kernel_root_dev = fopen(KERNEL_REAL_ROOT_DEV_FILE, "r")) != NULL) {
  	if(fgets(str_root_dev, 5, kernel_root_dev) != 0) {
			/* the kernel prints the real root device in decimal */
			fclose(kernel_root_dev);
			return (unsigned int) atoi(str_root_dev);
		} else {
			fclose(kernel_root_dev);
			printf("Unable to read from %s\n", KERNEL_REAL_ROOT_DEV_FILE);
			return -2;
 	  }
 		fclose(kernel_root_dev);
	} else {
		printf("Unable to open %s\n", KERNEL_REAL_ROOT_DEV_FILE);
   	return -1;
  }
}

/* set_real_root_dev returns 0 if the value could successfully be written
 * to the kernel and a value < 0 on error.
 * root_dev can be any value from 0 to 255
 */
int set_real_root_dev(const unsigned int root_dev) {
  FILE *kernel_root_dev;
  char str_root_dev[5];

  /* sanity check */
/*  if(root_dev > 0xff) {
  	printf("The root device number can only be from 0 to 255.\n");
  	return -3;
  } */

  /* the kernel can take the parameter in a decimal value */
  sprintf(str_root_dev, "%d", root_dev);

  if( (kernel_root_dev = fopen(KERNEL_REAL_ROOT_DEV_FILE, "w")) != NULL) {
  	if(fputs(str_root_dev, kernel_root_dev) != EOF) {
			fclose(kernel_root_dev);
			return 0;
		} else {
			fclose(kernel_root_dev);
			printf("Unable to write to %s\n", KERNEL_REAL_ROOT_DEV_FILE);
			return -2;
 	  }
 		fclose(kernel_root_dev);
	} else {
		printf("Unable to open %s\n", KERNEL_REAL_ROOT_DEV_FILE);
   	return -1;
  }
}

/* get_root_from_cmdline return 0 on success and the root device as string
 * in root_dev. On error, the return value is < 0.
 */
int get_root_from_cmdline(char *root_dev) {
  FILE *kernel_cmdline;
  char *root_dev_start, cmdline[MAX_CMDLINE_LENGTH];
  int i, len;

 /* parse the kernel commandline for an occurance of a root statement */
  if( (kernel_cmdline = fopen(KERNEL_CMDLINE, "r")) == NULL) {
	  if(fgets(cmdline, MAX_CMDLINE_LENGTH, kernel_cmdline) == NULL) {
    	printf("Unable to read from %s !\n", KERNEL_CMDLINE);
    	fclose(kernel_cmdline);
  	  return -2;
	  }
  	fclose(kernel_cmdline);
  	
	  if( (root_dev_start = strstr(cmdline, "root=")) != NULL) {
	    len = strlen(root_dev_start);
  	  i = 5;
	    while(i < len && root_dev_start[i] != ' ') {
    	  root_dev[i-5] = root_dev_start[i];
  	    i++;
	    }
    	root_dev[i-5] = 0;
    	return 0;
	  } else {
  	  printf("Warning: Could not find a root device parameter on kernel command line.\n");
  	  return -3;
	  }
	} else {
	  printf("Unable to open %s !\n", KERNEL_CMDLINE);
    return -1;
 	}
}
